% Copyright (C) 2001 Michel Juillard
%
function dr=olr2(dr,olr_inst,bet,obj_var,W)

global jacobia_ iy_ ykmin_ ykmax_ gstep_ exo_nbr endo_nbr
global ex_ valf_ it_ exe_ xkmin_ xkmax_ ys_ stdexo_
global fname_ means_ Sigma_e_ lgy_
global eigval options_

options_ = set_default_option(options_,'loglinear',0);

xlen = xkmax_ + xkmin_ + 1;
klen = ykmin_ + ykmax_ + 1;
iyv = iy_';
iyv = iyv(:);
iyr0 = find(iyv) ;
it_ = ykmin_ + 1 ;

inst_nbr = size(olr_inst,1);
inst_i = zeros(inst_nbr,1);
for i=1:inst_nbr
  k = strmatch(olr_inst(i,:),lgy_,'exact');
  if isempty(k)
    error(sprintf('OLR_INST %s isn''t a declared variable'));
  else
    inst_i(i) = k;
  end
end

if ykmax_ == 0
  error ('OLR : No forward variable: no point in using OLR') ;
end

if find(any(iy_([1:ykmin_ ykmin_+2:ykmax_],inst_i),2))
  error('OLR: instruments can only appear at the current period');
end

non_inst_i = setdiff([1:endo_nbr],inst_i);
iy1_ = iy_(:,non_inst_i);
endo_nbr_1 = endo_nbr - inst_nbr;

if exo_nbr == 0
  exe_ = [] ;
end

if ~ all(iy_(ykmin_+1,:) > 0)
  error ('Error in model specification: some variables don"t appear as current') ;
end

if xlen > 1
  error (['SS: stochastic exogenous variables must appear only at the' ...
	  ' current period. Use additional endogenous variables']) ;
end

  
dr=set_state_space(dr);
kstate = dr.kstate;
kad = dr.kad;
kae = dr.kae;
nstatic = dr.nstatic;
nfwrd = dr.nfwrd;
npred = dr.npred;
nboth = dr.nboth;
order_var = dr.order_var;
nd = size(kstate,1);
stat_var = order_var(1:nstatic);
stat_var = setdiff(stat_var,inst_i);
% static variables in objective function
[stat_obj_var] = intersect(obj_var,stat_var);
n_stat_obj_var = length(stat_obj_var);
pred_var = find(any(iy_(1:ykmin_,:),1))';
pred_var = [stat_obj_var; pred_var];
npred = length(pred_var)-nboth;
nstatic = length(stat_var);
nstatic1 = nstatic+inst_nbr;

endo_nbr_1 = endo_nbr-inst_nbr;
sdyn = endo_nbr - nstatic1;

order_var = [ inst_i; setdiff(stat_var,inst_i); order_var(nstatic1+1:end)];

% building QQ, RR and UU
iq = [];
iu = [];
io1 = [];
io2 = [];
for i=1:length(obj_var)
  i1 = find(obj_var(i)==order_var);
  if i1 > inst_nbr
    iq = [iq; i1-inst_nbr];
    io1 = [io1; obj_var(i)];
  else
    iu = [iu; i1];
    io2 = [io2; obj_var(i)];
  end
end
QQ = zeros(sdyn,sdyn);
QQ(iq,iq) = W(io1,io1);
RR = zeros(sdyn,inst_nbr);
RR(iq,iu) = W(io1,io2);
UU = zeros(inst_nbr,inst_nbr);
UU(iu,iu) = W(io2,io2);

tempex = ex_;

it_ = ykmin_ + 1;
z = repmat(dr.ys,1,klen);
z = z(iyr0) ;
jacobia_=real(jacob_a('ff1_',[z; exe_])) ;

ex_ = tempex ;
tempex = [];

nz = size(z,1);
k1 = iy_(find([1:klen] ~= ykmin_+1),:);
b = jacobia_(1:endo_nbr_1,iy_(ykmin_+1,order_var(inst_nbr+1:end)));
a = b\jacobia_(1:endo_nbr_1,nonzeros(k1')); 
if any(isinf(a(:)))
  error('OLR: the model doesn''t determine current variables uniquely')
end
if exo_nbr
  fu = b\jacobia_(1:endo_nbr_1,nz+1:end);
end
% instruments' effects
b = b\jacobia_(1:endo_nbr_1,iy_(ykmin_+1,inst_i));

% buildind D and E
sdyn1 = endo_nbr-nstatic-inst_nbr;          % number of Lagrange multipliers
nd1 = nd+2*sdyn1+inst_nbr;
d = zeros(nd1,nd1) ;
e = d ;

% variables order:
% z(+1) = [y(+1)' lambda(+1)' u' y' lambda']
%model dynamics
k = find(kstate(:,2) >= ykmin_+2 & kstate(:,3));
d(1:sdyn1,k) = a(nstatic+1:end,kstate(k,3)) ;
k1 = find(kstate(:,2) == ykmin_+2);
a1 = eye(sdyn1);
e(1:sdyn1,k1) =  -a1(:,kstate(k1,1)-nstatic1);
k2 = find(kstate(:,2) <= ykmin_+1 & kstate(:,4));
e(1:sdyn1,k2+sdyn1+inst_nbr) = -a(nstatic+1:end,kstate(k2,4)) ;
k3 = find(kstate(:,2) == ykmin_+1);
k3 = k3(~ismember(kstate(k3,1),kstate(k1,1)));
d(1:sdyn1,k3+sdyn1+inst_nbr) = a1(:,kstate(k3,1)-nstatic1);
k4 = sum(kstate(:,2) >= ykmin_+2);
e(1:sdyn1,k4+sdyn1+[1:inst_nbr]) = -b;

%first order condition from Lagrangian with respect to y
d(sdyn1+[1:sdyn1],nfwrd+nboth+[1:npred+nboth]) = a(nstatic+1:end,kstate(k2,4))';
e(sdyn1+[1:sdyn1],nfwrd+nboth+sdyn1+[1:inst_nbr]) = -RR;
d(sdyn1+[1:sdyn1],k3+sdyn1+inst_nbr) = 2*QQ(:,kstate(k3,1)-nstatic1);
d(sdyn1+[1:sdyn1],nfwrd+sdyn1+inst_nbr+npred+2*nboth+[1:sdyn1]) = eye(sdyn1);
e(sdyn1+[1:sdyn1],nfwrd+sdyn1+inst_nbr+npred+2*nboth+[npred+1:sdyn1]) = -a(nstatic+1:end,kstate(k,3))';
e(sdyn1+[1:sdyn1],k1) =  -2*QQ(:,kstate(k1,1)-nstatic1);

%first order condition from Lagrangian with respect to u
d(2*sdyn1+[1:inst_nbr],k3+sdyn1+inst_nbr) = RR(kstate(k3,1)-nstatic1,:)';
e(2*sdyn1+[1:inst_nbr],nfwrd+nboth+sdyn1+[1:inst_nbr]) = -2*UU;
d(2*sdyn1+[1:inst_nbr],nfwrd+sdyn1+inst_nbr+npred+2*nboth+[1:sdyn1]) = b';
e(2*sdyn1+[1:inst_nbr],k1) =  -RR(kstate(k1,1)-nstatic1,:)';

%auxiliary equations
d(2*sdyn1+inst_nbr+[1:sdyn1],nfwrd+sdyn1+inst_nbr+npred+2*nboth+[1:sdyn1]) = eye(sdyn1);
e(2*sdyn1+inst_nbr+[1:sdyn1],nfwrd+nboth+[1:sdyn1]) = eye(sdyn1);
if ~isempty(kad)
  for j = 1:size(kad,1)
    d(3*sdyn1+inst_nbr+j,nfwrd+sdyn1+inst_nbr+kad(j)) = 1 ;
    e(3*sdyn1+inst_nbr+j,kae(j)) = 1 ;
  end
end


options_ = set_default_option(options_,'qz_criterium',1.000001);
if ~exist('mjdgges')
  % using Chris Sim's routines
  use_qzdiv = 1;
  [ss,tt,qq,w] = qz(e,d);
  [tt,ss,qq,w] = qzdiv(options_.qz_criterium,tt,ss,qq,w);
  ss1=diag(ss);
  tt1=diag(tt);
  warning_state = warning;
  warning off;
  eigval = ss1./tt1 ;
  warning warning_state;
  nba = nnz(abs(eigval) > options_.qz_criterium);
else
  use_qzdiv = 0;
  [ss,tt,w,sdim,eigval,info] = mjdgges(e,d,options_.qz_criterium);
  if info & info ~= nd1+2;
    error(['ERROR' info ' in MJDGGES.DLL']);
  end
  nba = nd1-sdim;
end

nyf = nfwrd+nboth+sdyn1+inst_nbr;

if nba ~= nyf;
  disp('WARNING: Blanchard-Kahn conditions are not satisfied. Run CHECK to learn more!');
  disp('Press any key to continue');
  pause
end

np = nd1 - nyf;
n2 = np + 1;
n3 = nyf;
n4 = n3 + 1;
% derivatives with respect to dynamic state variables
% forward variables
gx = -w(1:n3,n2:nd1)'\w(n4:nd1,n2:nd1)';
% predetermined variables
hx = w(1:n3,1:np)'*gx+w(n4:nd1,1:np)';
hx = (tt(1:np,1:np)*hx)\(ss(1:np,1:np)*hx);

if use_qzdiv
  gx = real(gx);
  hx = real(hx);
end

% including Lagrange multipliers in lgy_, order_var and kstate
for i=1:sdyn1
  temp = ['mult_' int2str(i)];
  lgy_ = strvcat(lgy_,temp);
end

order_var = [order_var; endo_nbr+[1:sdyn1]'];

kstate = [kstate(1:nfwrd+nboth,:);zeros(sdyn1+inst_nbr,4);kstate(nfwrd+nboth+1:end,:);zeros(sdyn1,4)];
k = find(kstate(:,2) <= ykmin_+1);
kstate(nfwrd+nboth+[1:sdyn1+inst_nbr],1:2) = [[endo_nbr+[1:sdyn1]'; ...
		    order_var(1:inst_nbr)] (ykmin_+2)*ones(sdyn1+inst_nbr,1)];
kstate(2*sdyn1+nboth+inst_nbr+[1:sdyn1],1:2) = [endo_nbr+[1:sdyn1]' ...
		    (ykmin_+1)*ones(sdyn1,1)];

dr.ghx = [hx; gx(nboth+sdyn1+1:end,:)];

%lead variables actually present in the model
% derivatives with respect to exogenous variables
if exo_nbr
  n1 = find(kstate(:,2) > ykmin_+1);
  ghu = -(d*[zeros(nd1,nboth+nfwrd+sdyn1+inst_nbr) [gx;eye(npred+nboth+sdyn1)]]-e* ...
	  [[eye(nboth+nfwrd+sdyn1+inst_nbr); zeros(npred+nboth+sdyn1, ...
						   nboth+nfwrd+sdyn1+inst_nbr)] zeros(nd1,npred+nboth+sdyn1)])\[fu; zeros(size(d,1)-size(fu,1),exo_nbr)];
end


dr.ghx = [gx(nboth+nfwrd+sdyn1+[1:inst_nbr],:); hx(1:npred,:); gx(1:nboth+npred+sdyn1,:)]; 
dr.ghu = ghu([nboth+nfwrd+sdyn1+[1:inst_nbr] nboth+nfwrd+sdyn1+ ...
	      inst_nbr+[1:npred] [1:nboth+npred+sdyn1]],:); 

% static variables
if nstatic > 0
  j3 = nonzeros(kstate(:,3));
  j4  = find(kstate(:,3))-np-nyf;
  temp = -a(1:nstatic,j3)*gx(j4,:)*hx;
  temp = temp + b(1:nstatic,:)*gx(nyf+1:nyf+inst_nbr,:);
  j5 = find(kstate(:,4));
  temp(:,j5) = temp(:,j5)-a(1:nstatic,nonzeros(kstate(:,4)));
  dr.ghx = [dr.ghx(1:inst_nbr,:); temp; dr.ghx(inst_nbr+1:end,:)];
  temp = -a(1:nstatic,j3)*gx(j4,:)*ghu(1:np+nyf,:);
  temp = temp + b(1:nstatic,:)*ghu(np+2*nyf+1:np+2*nyf+inst_nbr,:);
  temp = temp + fu(1:nstatic,:);
  dr.ghu = [dr.ghu(1:inst_nbr,:); temp; dr.ghu(inst_nbr+1:end,:)];
  temp = [];
end

dr.ys = [dr.ys; zeros(sdyn1,1)];
dr.nstatic = nstatic1;
dr.npred = npred+nboth+sdyn1;
dr.kstate = kstate;
dr.order_var = order_var;
endo_nbr = endo_nbr+sdyn1;
